/*
** Module   : MenuItemGuiImpl.java
** Abstract : GUI client implementation of MenuItem.
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150525 Created initial version.
*/
package com.goldencode.p2j.ui.client.gui;

import java.awt.event.MouseEvent;

import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.*;
import com.goldencode.p2j.ui.client.event.*;
import com.goldencode.p2j.ui.client.gui.GuiFontResolver.FontCache;
import com.goldencode.p2j.ui.client.gui.driver.*;
import com.goldencode.p2j.ui.client.widget.*;

/**
 * <code>MenuItem</code> GUI client implementation class. 
 */
public class MenuItemGuiImpl
extends MenuItem<GuiOutputManager>
{
   /** Font resolver */
   private GuiFontResolver gf;

   /** Screen driver */
   private GuiDriver gd;

   /** Check mark color. */
   private ColorRgb chkMarkColor;
   
   /** shadow color. */
   private ColorRgb rule3DShadow;
   
   /** highlighted color */
   private ColorRgb highlightedColor;
   
   /** True when mouse is over the flat button.  */
   private boolean mouseEntered = false;

   /** True when mouse is over the flat button.  */
   private boolean mouseDragged = false;
   
   /** Pressed button state flag. */
   private boolean pressed = false;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public MenuItemGuiImpl(WidgetId id, MenuItemConfig cfg)
   {
      super(id, cfg);
   }

   /**
    * Replacement API for the logic which should exist in the widget's c'tor, but the logic 
    * can't be executed as it depends on the widget being registered into the official registry.
    * <p>
    * This API is called after the widget is created and added to the registry.
    * 
    * @param    id
    *           The widget's ID.
    * @param    cfg
    *           The config used to initialize this widget.
    */
   @Override
   public void initialize(WidgetId id, MenuItemConfig cfg)
   {
      super.initialize(id, cfg);

      gd = (GuiDriver) OutputManager.getDriver();
      
      // the label's window is resolved from the frame... the label is not necessarily attached
      // to the frame at the time of this call, just the frame ID is known

      Widget<?> parent = ThinClient.getInstance().getWidget(cfg.frameId);
      Window<?> window = parent.window();
      
      gf = new GuiFontResolver(window, gd, config(), false);

      chkMarkColor       = gd.getSysColor(SysColor.COLOR_WINDOWTEXT);
      rule3DShadow       = gd.getSysColor(SysColor.COLOR_3DSHADOW);
      highlightedColor   = gd.getSysColor(SysColor.COLOR_MENUHILIGHT);
      //updateSize();
   }
   
   /**
    * Get widget dimension.
    * 
    * @return  widget dimension.
    */
   @Override
   public Dimension dimension()
   {
      return new Dimension(width(), 1);
   }

   /**
    * Draw widget.
    */
   @Override
   public void draw()
   {
      final GuiColorResolver gce = new GuiColorResolver(gd, config(), 
               SysColor.COLOR_MENU, 
               SysColor.COLOR_BTNTEXT);
      gce.colors();
      
      final FontCache fc = gf.font();
      
      final NativePoint p = physicalLocation();
      final NativeDimension dim = physicalDimension();
      NativeRectangle clip = clipRectangle(p, dim);
      gd.draw(p, clip, new Runnable()
      {
         @Override
         public void run()
         {
            // rectangle around
            
            drawTitleRectangle(gce, dim);
            
            // SKIP or RULE
            if (isSkip())
            {
               return; 
            }

            // RULE
            if (isRule())
            {
               gd.setColor(rule3DShadow);
               gd.drawLine(2, dim.height / 2, dim.width - 2, dim.height / 2);
               return;
            }
            
            // NORMAL
            Window<?> window = Window.resolveWindow(MenuItemGuiImpl.this);
            int fh = FontManager.getFontHeight(window, gf.resolveFontNum(), gd);
            
            gd.setColor(gce.fgColor);
            gd.setGuiFont(fc.font);
            gd.setFontStyle(fc.font.style);
            
            NativePoint tp = new NativePoint(7, 2 + fh);
            
            if (config().toggleBox && config().checked)
            {
               // TODO test: what if toggleBox = true and checked = false
               tp.y -= 4 + fh/2;
               ToggleBoxGuiImpl.drawCheckmark(tp, chkMarkColor);
               tp.x += 13;
               tp.y += 4 + fh/2;
            }
            
            // disabled menu-item is drawn differently: 
            /* TODO if (isInitDisabled())
            {
               // remove mnemonic if specified 
               title = title.replaceAll("&", "");
               
               if (!hasFocus())
               {
                  gd.drawString(title, p.x, p.y);
                  return;
               }
            }*/
            
            String title = config().title;
            
            // if & is not specified use the first letter as mnemonic
            int ampIdx = title.indexOf('&');
            boolean amp = ampIdx != -1; 
            if (!amp) ampIdx = config().toggleBox ? 1 : 0;

            String leftPart = title.substring(0, ampIdx);
            gd.drawString(leftPart, tp.x, tp.y);
            
            tp.x += FontManager.getTextWidth(leftPart, window, gf.resolveFontNum(), gd);
            
            // "eat" the ampersand
            if (amp) ampIdx++;
            
            String mnem = title.substring(ampIdx, ampIdx + 1);
            gd.setFontStyle(FontStyle.UNDERLINE);
            gd.drawString(mnem, tp.x, tp.y);

            tp.x += FontManager.getTextWidth(mnem, window, gf.resolveFontNum(), gd);
            
            gd.setFontStyle(fc.font.style);            
            gd.drawString(title.substring(ampIdx + 1), tp.x, tp.y);
         }

         /**
          * Draws rectangle around title.
          * 
          * @param  gce
          *         Color resolver.
          *  
          * @param  dim
          *         Dimension of this widget. 
          */
         private void drawTitleRectangle(final GuiColorResolver gce, final NativeDimension dim)
         {
            Container<?> parent = MenuItemGuiImpl.this.parent();
            if (parent instanceof Menu && !((Menu<?>)parent).config().popupOnly) // menubar
            {
               if (MenuItemGuiImpl.this.pressed)
               {
                  gd.setColor(rule3DShadow);
                  gd.draw3DRect(2, 2, dim.width - 2, dim.height - 2, false);
               } 
               else if (MenuItemGuiImpl.this.mouseEntered)
               {
                  gd.setColor(rule3DShadow);
                  gd.draw3DRect(2, 2, dim.width - 2, dim.height - 2, true);
               } 
               else 
               {
                  gd.setColor(gce.bgColor);
                  gd.drawRect(2, 2, dim.width - 2, dim.height - 2);
               }
            }
            else
            {
               if (MenuItemGuiImpl.this.mouseEntered)
               {
                  Window<?> wnd = Window.resolveWindow(MenuItemGuiImpl.this);
                  int width = wnd.screen().coordinates().widthToNative(parent().width());
                  gd.setColor(highlightedColor);
                  gd.fillRect(2, 2, width - 2, dim.height - 2);
               } 
            }
         }
      });
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      Window<?> window = Window.resolveWindow(this);
      int h = FontManager.getFontHeight(window, gf.resolveFontNum(), gd) + 5 + 5;
      return window.screen().coordinates().heightFromNative(h);
   }

   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      if (config().fixedWidth)
      {
         return config().widthChars;
      }
      
      MenuItemConfig cfg = config();
      
      if (!isNormal())
      {
         // set width to longest neighbor's width.
         double width = 0;
         for (Widget<?> wdg : parent().widgets())
         {
            // prevent infinite loop
            if (wdg.equals(this))
            {
               continue;
            }
            
            // prevent infinite loop
            if (wdg instanceof MenuItem && !((MenuItem<?>)wdg).isNormal())
            {
               continue;
            }
            
            double w = wdg instanceof SubMenu ? ((SubMenu<?>) wdg).titleWidth() : wdg.width();
            
            if (width < w)
            {
               width = w;
            }
         }
         
         return width;
      }
      
      String title = cfg.title; 
      
      if (title.contains("&"))
      {
         int amp = title.indexOf("&");
         title = title.substring(0, amp) + title.substring(amp + 1);
      }
      
      Window<?> window = Window.resolveWindow(this);
      // method calls server-side to get legacy text metrics, so current changes could be sent.
      int nativeWidth = FontManager.getTextWidth(title, window, gf.resolveFontNum(), gd);
      if (cfg.toggleBox)
         nativeWidth += 13;
      nativeWidth += 5 + 5; // padding
      
      return window.screen().coordinates().widthFromNative(nativeWidth);
   }

   /**
    * Get widget minimum size.
    * 
    * @return  widget minimum size.
    */
   @Override
   public Dimension minimumSize()
   {
      return dimension();
   }

   /**
    * Notification of a mouse clicked event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseClicked(MouseEvent e)
   {
      // and simulate ENTER key to select it
      KeyInput ki = new KeyInput(this, EventType.KEY_PRESSED, Key.VK_ENTER);
      EventManager.postEvent(ki);
   }
   
   /**
    * Notification of a mouse entered event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseEntered(MouseEvent e)
   {
      boolean oldVal = mouseEntered; 
      mouseEntered = true;
      
      if (!oldVal)
      {
         requestFocus();
         parent().repaint();
      }
   }

   /**
    * Notification of a mouse exited event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseExited(MouseEvent e)
   {
      boolean oldVal = mouseEntered; 
      mouseEntered = false;
      
      if (oldVal)
      {
         parent().repaint();
      }
      
      mouseDragged = false;
   }

   /**
    * Notification of a mouse pressed event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mousePressed(MouseEvent e)
   {
      boolean oldVal = pressed; 
      pressed = true;
      
      if (!oldVal)
      {
         parent().repaint();
      }
   }

   /**
    * Notification of a mouse released event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseReleased(MouseEvent e)
   {
      boolean oldVal = pressed; 
      pressed = false;
      
      if (oldVal)
      {
         parent().repaint();
      }
   }

   /**
    * Notification of a mouse dragged event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseDragged(MouseEvent e)
   {
      // get coords of the drag event related to button
      NativePoint mp = translate(this, e);
      // get button location and size
      NativePoint p = physicalLocation();
      NativeDimension d = physicalDimension();
      // this can be rare case when we leave the button with holded mouse key, the button must
      // be released, the mouseExited event will not be delivered to button until drag button
      // released even if the pointer is far away from button accepted press event
      if ((mp.x < p.x || mp.x > (p.x + d.width) || mp.y < p.y || mp.y > (p.y + d.height)) &&
          pressed)
      {
         pressed = false;
         mouseEntered = false;
         parent().repaint();
      }
      
      mouseDragged = true;
   }

   /**
    * Notification of a mouse moved event occurred for this widget.  No-op by default.
    * 
    * @param    e
    *           The mouse event.
    */
   @Override
   public void mouseMoved(MouseEvent e)
   {
      // TODO
   }
}
